// Version information for the "groupsock" library
// Copyright (c) 1996-2016 Live Networks, Inc.  All rights reserved.

#ifndef _GROUPSOCK_VERSION_HH
#define _GROUPSOCK_VERSION_HH

#define GROUPSOCK_LIBRARY_VERSION_STRING	"2016.08.07"
#define GROUPSOCK_LIBRARY_VERSION_INT		1470528000

#endif
